﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;
using SharpDX.DirectInput;

namespace GameUtils.Input.DefaultDevices
{
    public class Mouse : InputDevice<MouseState>
    {
        DirectInput input;
        SharpDX.DirectInput.Mouse mouse;
        AutoResetEvent eventHandle;
        Thread t;
        bool capture;
        IntPtr surfaceHandle;
        LinkedList<IInputListener<MouseState>> listeners;
        
        [DllImport("user32.dll")]
        private static extern bool ScreenToClient(IntPtr hWnd, ref POINT lpPoint);

        [StructLayout(LayoutKind.Sequential)]
        private struct POINT
        {
            public int X;
            public int Y;
        }

        public Mouse()
        {
            listeners = new LinkedList<IInputListener<MouseState>>();
        }

        public override ListenerHandle<MouseState> RegisterListener(IInputListener<MouseState> listener)
        {
            LinkedListNode<IInputListener<MouseState>> node = listeners.AddLast(listener);
            return new ListenerHandle<MouseState>(listeners, node);
        }

        public override void Initialize(Control surface)
        {
            surfaceHandle = surface.Handle;
            input = new DirectInput();
            mouse = new SharpDX.DirectInput.Mouse(input);
            mouse.SetCooperativeLevel(surface, CooperativeLevel.Background | CooperativeLevel.NonExclusive);
        }

        private Point ScreenPointToSurfacePoint(Point p)
        {
            POINT pt = new POINT();
            pt.X = p.X;
            pt.Y = p.Y;
            ScreenToClient(surfaceHandle, ref pt);
            return new Point(pt.X, pt.Y);
        }

        public override void BeginCapture()
        {
            eventHandle = new AutoResetEvent(true);
            mouse.SetNotification(eventHandle);
            mouse.Acquire();
            capture = true;
            t = new Thread(() =>
            {
                do
                {
                    SharpDX.DirectInput.MouseState stateDI = mouse.GetCurrentState();
                    MouseButtons buttons = MouseButtons.None;
                    if (stateDI.Buttons[0]) buttons |= MouseButtons.Left;
                    if (stateDI.Buttons[2]) buttons |= MouseButtons.Right;
                    if (stateDI.Buttons[1]) buttons |= MouseButtons.Middle;
                    Point cursorPos = ScreenPointToSurfacePoint(Cursor.Position);
                    MouseState state = new MouseState(buttons, stateDI.X, stateDI.Y, cursorPos.X, cursorPos.Y, stateDI.Z);
                    foreach (IInputListener<MouseState> listener in listeners)
                    {
                        listener.OnInputReceived(state);
                    }
                    eventHandle.WaitOne();
                } while (capture);
            });
            t.Start();
        }

        public override void EndCapture()
        {
            capture = false;
            eventHandle.Set();
            t.Join();
            mouse.Unacquire();
            eventHandle.Dispose();
            eventHandle = null;
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                listeners = null;
                if (eventHandle != null) eventHandle.Dispose();
            }

            mouse.Dispose();
            input.Dispose();
        }
    }
}
