﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using GameUtils.Graphics;
using GameUtils.Graphics.Data;
using GameUtils.Math;

namespace GameUtils.Renderers.Gdi
{
    internal class GdiRadialGradientBrush : RadialGradientBrushData, IGdiBrush
    {
        GraphicsPath path;
        PathGradientBrush brush;
        GradientStop[] stops;
        Vector2 center;
        float radiusX;
        float radiusY;
        Vector2 originOffset;
        float opacity;
        Matrix matrix;

        public Brush Brush
        {
            get { return brush; }
        }

        protected override GradientStop[] GradientStops
        {
            get { return stops; }
            set
            {
                if (value == null)
                    throw new ArgumentNullException("value");

                stops = value;
                brush.Dispose();
                path.Dispose();
                brush = this.CreateBrush(out path);
            }
        }

        protected override Vector2 Center
        {
            get { return center; }
            set
            {
                center = value;
                brush.Dispose();
                path.Dispose();
                brush = this.CreateBrush(out path);
            }
        }

        protected override float RadiusX
        {
            get { return radiusX; }
            set
            {
                radiusX = value;
                brush.Dispose();
                path.Dispose();
                brush = this.CreateBrush(out path);
            }
        }

        protected override float RadiusY
        {
            get { return radiusY; }
            set
            {
                radiusY = value;
                brush.Dispose();
                path.Dispose();
                brush = this.CreateBrush(out path);
            }
        }

        protected override Vector2 OriginOffset
        {
            get { return originOffset; }
            set
            {
                originOffset = value;
                brush.CenterPoint = GdiHelper.Vector2ToPointF(center + originOffset);
            }
        }

        protected override float Opacity
        {
            get { return opacity; }
            set
            {
                if (value > 1)
                    opacity = 1;
                else if (value < 0)
                    opacity = 0;
                else
                    opacity = value;
                brush.Dispose();
                path.Dispose();
                brush = this.CreateBrush(out path);
            }
        }

        private PathGradientBrush CreateBrush(out GraphicsPath path)
        {
            path = new GraphicsPath();
            path.AddEllipse(new RectangleF(center.X - radiusX, center.Y - radiusY, 2 * radiusX, 2 * radiusY));
            var b = new PathGradientBrush(path);
            b.CenterPoint = GdiHelper.Vector2ToPointF(center + originOffset);
            var blend = new ColorBlend(stops.Length);
            for (int i = 0; i < stops.Length; i++)
            {
                var color = stops[i].Color;
                color.A *= opacity;
                blend.Colors[i] = GdiHelper.ConvertColor(color);
                blend.Positions[i] = stops[i].Position;
            }
            b.InterpolationColors = blend;
            if (matrix != null) b.Transform = matrix;
            return b;
        }

        internal GdiRadialGradientBrush(GradientStop[] stops, Vector2 center, float radiusX, float radiusY)
        {
            this.stops = stops;
            this.center = center;
            this.radiusX = radiusX;
            this.radiusY = radiusY;
            opacity = 1;
            brush = this.CreateBrush(out path);
        }

        protected override void SetTransform(Matrix2x3 matrix)
        {
            var gdiMatrix = new Matrix(matrix[0, 0], matrix[1, 0], matrix[0, 1], matrix[1, 1], matrix.OffsetX, matrix.OffsetY);
            brush.Transform = gdiMatrix;
            if (this.matrix != null)
                this.matrix.Dispose();
            this.matrix = gdiMatrix;
        }

        protected override void ResetTransform()
        {
            brush.ResetTransform();
            if (matrix != null)
                matrix.Dispose();
            matrix = null;
        }

        protected override void Dispose(bool disposing)
        {
            brush.Dispose();
            path.Dispose();
            if (matrix != null) matrix.Dispose();
        }
    }
}
