﻿using GameUtils.Graphics.Data;
using GameUtils.Math;

namespace GameUtils.Graphics
{
    public abstract class Brush<T> : Resource<T> where T : BrushData
    {
        float opacity;
        Matrix2x3 matrix;
        bool transformed;
        bool opacityChanged;
        bool transformChanged;
        readonly object locker;

        /// <summary>
        /// A value between 1 and 0 indicating the transparency of this brush.
        /// </summary>
        public float Opacity
        {
            get { return opacity; }
            set
            {
                lock (locker)
                {
                    opacity = MathHelper.Clamp(value, 0, 1);
                    opacityChanged = true;
                }
            }
        }

        protected Brush()
        {
            opacity = 1;
            locker = new object();
        }

        protected override void DataCreated(T data)
        {
            lock (locker)
            {
                data.Opacity = opacity;
                if (transformed) data.SetTransform(matrix);
                opacityChanged = false;
                transformChanged = false;
            }
        }

        protected override void ApplyChanges()
        {
            lock (locker)
            {
                if (opacityChanged)
                {
                    Data.Opacity = opacity;
                    opacityChanged = false;
                }
                if (transformChanged)
                {
                    if (transformed)
                        Data.SetTransform(matrix);
                    else
                        Data.ResetTransform();
                    transformChanged = false;
                }
            }
        }

        /// <summary>
        /// Applies a transformation to this brush.
        /// </summary>
        public void SetTransform(Matrix2x3 matrix)
        {
            lock (locker)
            {
                this.matrix = matrix;
                transformed = true;
                transformChanged = true;
            }
        }

        /// <summary>
        /// Resets the transformation on this brush.
        /// </summary>
        public void ResetTransform()
        {
            lock (locker)
            {
                matrix = Matrix2x3.Identity;
                transformed = false;
                transformChanged = true;
            }
        }
    }
}
