﻿using System;
using GameUtils.Graphics.Data;

namespace GameUtils.Graphics
{
    public class Font : Resource<FontData>
    {
        string familyName;
        float size;
        FontStyle style;
        FontWeight weight;
        bool familyNameChanged;
        bool sizeChanged;
        bool styleChanged;
        bool weightChanged;
        readonly object locker;

        /// <summary>
        /// The family name of this font.
        /// </summary>
        public string FamilyName
        {
            get { return familyName; }
            set
            {
                if (value == null)
                    throw new ArgumentNullException("value");

                lock (locker)
                {
                    familyName = value;
                    familyNameChanged = true;
                }
            }
        }

        /// <summary>
        /// The size of this font in pixels.
        /// </summary>
        public float Size
        {
            get { return size; }
            set
            {
                if (value < 0)
                    throw new ArgumentOutOfRangeException("value");

                lock (locker)
                {
                    size = value;
                    sizeChanged = true;
                }
            }
        }

        /// <summary>
        /// The style of this font.
        /// </summary>
        public FontStyle Style
        {
            get { return style; }
            set
            {
                lock (locker)
                {
                    style = value;
                    styleChanged = true;
                }
            }
        }

        /// <summary>
        /// The weight of this font.
        /// </summary>
        public FontWeight Weight
        {
            get { return weight; }
            set
            {
                lock (locker)
                {
                    weight = value;
                    weightChanged = true;
                }
            }
        }

        /// <summary>
        /// Creates a new font.
        /// </summary>
        public Font(string familyName, float size, FontStyle style = FontStyle.Normal, FontWeight weight = FontWeight.Normal)
        {
            this.familyName = familyName;
            this.size = size;
            this.style = style;
            this.weight = weight;
            locker = new object();
        }

        protected override void ApplyChanges()
        {
            lock (locker)
            {
                if (familyNameChanged)
                {
                    Data.FamilyName = familyName;
                    familyNameChanged = false;
                }
                if (sizeChanged)
                {
                    Data.Size = size;
                    sizeChanged = false;
                }
                if (styleChanged)
                {
                    Data.Style = style;
                    styleChanged = false;
                }
                if (weightChanged)
                {
                    Data.Weight = weight;
                    weightChanged = false;
                }
            }
        }

        protected override FontData CreateData(Factory factory)
        {
            lock (locker)
            {
                familyNameChanged = false;
                sizeChanged = false;
                styleChanged = false;
                weightChanged = false;
                return factory.CreateFont(familyName, size, style, weight);
            }
        }
    }
}
