﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using GameUtils.Graphics.Data;
using GameUtils.Math;

namespace GameUtils.Graphics
{
    /// <summary>
    /// The renderer is a singletone component and performs basic drawing operations.
    /// </summary>
    public abstract class Renderer : IEngineComponent, IDisposable
    {
        /// <summary>
        /// Creates a renderer.
        /// </summary>
        /// <remarks>Use this method to directly choose the renderer.</remarks>
        public static Renderer Create<T>(Control surface, PresentationOption presentationOption = PresentationOption.Limited, AntiAliasMode antiAliasMode = AntiAliasMode.AntiAlias) where T : Renderer, new()
        {
            T renderer = new T();
            if (renderer.PlatformVersion > Environment.OSVersion.Version)
                throw new NotSupportedException("This renderer is not supported by the current platform.");
            renderer.Initialize(surface, presentationOption, antiAliasMode);
            return renderer;
        }

        /// <summary>
        /// Creates a renderer.
        /// </summary>
        /// <remarks>Use this method if you want the engine to choose the renderer.</remarks>
        public static Renderer IntelligentSelect(IEnumerable<Renderer> renderers, Control surface, PresentationOption presentationOption = PresentationOption.Limited, AntiAliasMode antiAliasMode = AntiAliasMode.AntiAlias)
        {
            Renderer renderer = renderers.OrderByDescending(item => item.PlatformVersion).FirstOrDefault(item => item.PlatformVersion <= Environment.OSVersion.Version);
            if (renderer == null)
                throw new NotSupportedException("None of the specified renderers is supported by the current platform.");
            renderer.Initialize(surface, presentationOption, antiAliasMode);
            return renderer;
        }

        bool IEngineComponent.IsCompatibleTo(IEngineComponent component)
        {
            return !(component is Renderer);
        }

        object IEngineComponent.Tag
        {
            get { return null; }
        }

        protected internal abstract Factory Factory { get; }

        /// <summary>
        /// The minimum platform version required by this renderer.
        /// </summary>
        public abstract Version PlatformVersion { get; }

        /// <summary>
        /// The bounds of the renderer's surface.
        /// </summary>
        public abstract Rectangle SurfaceBounds { get; }

        protected abstract void Initialize(Control surface, PresentationOption presentationOption, AntiAliasMode antiAliasMode);

        /// <summary>
        /// Extracts the underlying data of a resource.
        /// </summary>
        /// <param name="resource">The resource to extract the data from.</param>
        protected T GetResourceData<T>(Resource<T> resource) where T : ResourceData
        {
            return resource.Data;
        }

        protected internal abstract void BeginRender();

        protected internal abstract void EndRender();

        public abstract void DrawLine<T>(Vector2 p1, Vector2 p2, Brush<T> brush, float strokeWidth = 1.0f) where T : BrushData;
        public abstract void DrawLine<T>(Vector2 p1, Vector2 p2, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle) where T : BrushData;

        public abstract void DrawEllipse<T>(Ellipse ellipse, Brush<T> brush, float strokeWidth = 1.0f) where T : BrushData;
        public abstract void DrawEllipse<T>(Ellipse ellipse, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle) where T : BrushData;
        public abstract void FillEllipse<T>(Ellipse ellipse, Brush<T> brush) where T : BrushData;

        public abstract void DrawRectangle<T>(Rectangle rect, Brush<T> brush, float strokeWidth = 1.0f) where T : BrushData;
        public abstract void DrawRectangle<T>(Rectangle rect, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle) where T : BrushData;
        public abstract void FillRectangle<T>(Rectangle rect, Brush<T> brush) where T : BrushData;

        public abstract void DrawText<T>(string text, Font font, Brush<T> brush, Rectangle rect, TextFormat format = null) where T : BrushData;
        public abstract void DrawText<T>(string text, Font font, Brush<T> brush, Vector2 pt) where T : BrushData;

        public abstract void DrawTexture(Texture texture, Rectangle destinationRect, Rectangle? sourceRect = null, float opactiy = 1.0f, InterpolationMode interpolationMode = InterpolationMode.HighQualityCubic);
        public abstract void DrawTexture(Texture texture, Vector2 pt, float opacity = 1.0f);

        public abstract void DrawGeometry<T>(Geometry geometry, Brush<T> brush, float strokeWidth = 1.0f) where T : BrushData;
        public abstract void DrawGeometry<T>(Geometry geometry, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle) where T : BrushData;
        public abstract void FillGeometry<T>(Geometry geometry, Brush<T> brush) where T : BrushData;

        public abstract void DrawPolygon<T>(Polygon polygon, Brush<T> brush, float strokeWidth = 1.0f) where T : BrushData;
        public abstract void DrawPolygon<T>(Polygon polygon, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle) where T : BrushData;
        public abstract void FillPolygon<T>(Polygon polygon, Brush<T> brush) where T : BrushData;

        /// <summary>
        /// Clears the surface with the specified Color.
        /// </summary>
        public abstract void Clear(Color4 color);

        /// <summary>
        /// Pushes a clipping region to the stack.
        /// </summary>
        public abstract void PushClip(Geometry geometry);

        /// <summary>
        /// Pushes a clipping rectangle to the stack.
        /// </summary>
        public abstract void PushClip(Rectangle rect);

        /// <summary>
        /// Removes the topmost clipping region from the stack.
        /// </summary>
        public abstract void PopClip();

        /// <summary>
        /// Sets the worls transformation matrix.
        /// </summary>
        public abstract void SetTransform(Matrix2x3 matrix);

        /// <summary>
        /// Resets the world transformationmatrix to the identity matrix.
        /// </summary>
        public abstract void ResetTransform();

        /// <summary>
        /// Calculates the size of a string with a specified font.
        /// </summary>
        public abstract Vector2 MeasureText(string text, Font font);

        private bool disposed;

        public void Dispose()
        {
            if (!disposed)
            {
                Dispose(true);
                GC.SuppressFinalize(this);

                disposed = true;
            }
        }

        protected virtual void Dispose(bool disposing)
        { }

        ~Renderer()
        {
            Dispose(false);
        }
    }
}
